/*
 * $Author: ac $
 * $Id: bflsort.c,v 1.2 1999/10/21 12:51:52 ac Exp $
 * $Date: 1999/10/21 12:51:52 $
 */
/*
 * $Log: bflsort.c,v $
 * Revision 1.2  1999/10/21 12:51:52  ac
 * Nouvelle version avec les librairies bdd104, abe, abv, abt.
 * Ajout de fpm_network pour accelerer l'optimisation.
 *
 * Revision 1.1.1.1  1998/10/21 16:32:38  alliance
 * fpmap import
 *
 */

/*----------------------------------------------------------------------*/
/*                               Includes                               */
/*----------------------------------------------------------------------*/

#include  BFL_H

/*----------------------------------------------------------------------*/
/*                        Constant declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                          Type declarations                           */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Stucture declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Variable declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                          Macro declarations                          */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                      Static function prototypes                      */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                    Definition of static functions                    */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                   Definition of exported functions                   */
/*----------------------------------------------------------------------*/

/*
 * Generic Heapsort.
 *
 * NAME
 *      bflsort - sorts an array
 * SYNOPSIS
 *      #include  BFL_H
 *
 *      void bflsort(bflptr_t base, bflsize_t n, bflsize_t size,
 *                      bflcompar_fn compar);
 * DESCRIPTION
 *      The \fBbflsort()\fP function sorts an array with \fBn\fP elements
 *      of size \fBsize\fP. The \fBbase\fP argument points to the start of
 *      the array.
 *
 *      The contents of the array are sorted in ascending order according
 *      to a comparison function pointed to by \fBcompar\fP, which is called
 *      with two arguments that point to the objects being compared.
 *
 *      The comparison function must return an integer less than, equal to,
 *      or greater than zero if the first argument is considered to be
 *      respectively less than, equal to, or greater than the second. If
 *      two members compare as equal, their order in the sorted array is
 *      undefined.
 *
 * RETURN VALUE
 *      The \fBbflsort()\fP function returns no value.
 */

#define swap(p1, p2, n) { \
	char *_p1; \
	char *_p2; \
	bflsize_t _n; \
	char _tmp; \
	\
	for (_p1 = (char *)p1, _p2 = (char *)p2, _n = n; _n-- > 0;) { \
		_tmp = *_p1; *_p1++ = *_p2; *_p2++ = _tmp; \
	} \
} \

/*
 * To avoid function calls in the inner loops, the code responsible for
 * constructing a heap from (part of) the array has been expanded inline.
 * It is possible to convert this common code to a function. The three
 * parameters base0, cmp and size are invariant - only the size of the
 * gap and the high bound of the array change. In phase 1, gap decreases
 * while hi is fixed. In phase 2, gap == size, and hi decreases. The
 * variables p, q, and g are only used in this common code.
 */

void
bflsort(base, n, size, compar)
    bflptr_t base;
    bflsize_t n;
    bflsize_t size;
    bflcompar_fn compar;
{
    bflptr_t p;
    bflptr_t q;
    bflptr_t base0;
    bflptr_t hi;
    bflsize_t gap, g;

    if (n < 2)
	return;

    base0 = base - size;	/* set up address of h[0] */

/* 
 * The gap is the distance, in bytes, between h[0] and h[i],
 * for some i. It is also the distance between h[i] and h[2*i];
 * that is, the distance between a node and its left child.
 * The initial node of interest is h[n/2] (the rightmost
 * interior node), so gap is set accordingly. The following is
 * the only multiplication needed.
 */

    gap = (n >> 1) * size;	/* initial gap is n/2*size */
    hi = base0 + gap + gap;	/* calculate address of h[n] */
    if (n & 1)
	hi += size;		/* watch out for odd arrays */

/* 
 * Phase 1: Construct heap from random data.
 *
 * For i = n/2 downto 2, ensure h[i] is greater than its
 * children h[2*i] and h[2*i+1]. By decreasing 'gap' at each
 * iteration, we move down the heap towards h[2]. The final step
 * of making h[1] the maximum value is done in the next phase.
 */

    for (; gap != size; gap -= size) {
	for (p = base0 + (g = gap); (q = p + g) <= hi; p = q) {
	    g += g;		/* double gap for next level */

	/* 
	 * Find greater of left and right children.
	 */
	    if (q != hi && (*compar) (q + size, q) > 0) {
		q += size;	/* choose right child */
		g += size;	/* follow right subtree */
	    }
	/* 
	 * Compare with parent.
	 */
	    if ((*compar) (p, q) >= 0)
		break;		/* order is correct */
	    swap(p, q, size);	/* swap parent and child */
	}
    }

/* 
 * Phase 2: Each iteration makes the first item in the
 * array the maximum, then swaps it with the last item, which
 * is its correct position. The size of the heap is decreased
 * each iteration. The gap is always "size", as we are interested
 * in the heap starting at h[1].
 */

    for (; hi != base; hi -= size) {
	p = base;		/* == base0 + size */
	for (g = size; (q = p + g) <= hi; p = q) {
	    g += g;
	    if (q != hi && (*compar) (q + size, q) > 0) {
		q += size;
		g += size;
	    }
	    if ((*compar) (p, q) >= 0)
		break;
	    swap(p, q, size);
	}
	swap(base, hi, size);	/* move largest item to end */
    }
}
